/* ------------------------------------------------------------------------- */
/* ezplot.c
 *
 * Copyright (C) 1998-2002 Dwight Barkley                 
 *
 * ------------------------------------------------------------------------- */

/* #define MY_KEY_EVENT_HANDLE */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include "ezplot.h"

/* ========================================================================= */

#define WINDOW_TITLE "EZ-Plot"

#define PLOTSIZE       400	/* plot size in pixels  */
#define WINX             1	/* window location, in pixels, from left    */
#define WINY             1	/* window location, in pixels, from top     */
#define BORDER          10	/* border width in pixels                   */
#define BACKGROUND       1	/* background color (0 BLACK or 1 WHITE)    */
#define LINE_WIDTH       2	/* line width, in pixels, of plotted lines  */

/* Set the event_mask to events you wish to handle.  It is Not
   necessary to change this unless you modify the code to add new
   event such as mouse events.  */
#define EVENT_MASK	(KeyPressMask | EnterWindowMask | ExposureMask)

/* Should be no need to change anything below */
#define MODE_SIMULATING  1
#define MODE_VIEWING     2

/* 
 * Private variables 
 * ----------------- */
static int plot_size;		/* plot size in pixels */
static int plot_position[2];	/* plot location in pixels */
static int verbose = 2;		/* verbosity level */
static float x_scale, x_offset;
static float u_scale, u_offset;
static float c1, c2;

static Display *theDisplay;
static Window theWindow;
static Pixmap theUPixmap;
static GC theGC;
static Colormap theColormap;
static int theScreen;
static int theDepth;
static int thePWidth;
static int thePHeight;
static unsigned long theBlackPixel;
static unsigned long theWhitePixel;
static unsigned long theRedPixel;
static unsigned long theGreenPixel;
static unsigned long theBluePixel;

static int ezmode = MODE_VIEWING;

/* 
 * Private functions 
 * ----------------- */
static void Set_color (char color);
static void InitX (void);
static void OpenWindow (int x, int y, int width, int height);
static void InitColors (void);
static void QuitX (void);
static int Convert_to_pixel (double x, double scale, double offset);

/* ========================================================================= */

void
EZplot_simple1 (double *u, int npoints)
{
  /* Plot a blue curve (x,u) on a white background */

  double *x = (double *) malloc ((unsigned) (npoints) * sizeof (double));
  int j;
  for (j = 0; j < npoints; j++)
    x[j] = j;

  EZplot_clear ('w');
  EZplot_line1 (x, u, npoints, 'b');
  EZplot_display ();

  free (x);
}

/* ========================================================================= */

void
EZplot_line1 (double *x, double *u, int npoints, char color)
{
  /* Draw a colored curve (x,u) to the pixmap */

  XPoint *thePoints;
  int j;

  if (npoints <= 1)
    return;

  /* Allocate thePoints and copy x and u to thePoints */
  thePoints = (XPoint *) malloc ((unsigned) (npoints) * sizeof (XPoint));
  for (j = 0; j < npoints; j++)
    {
      thePoints[j].x = (short) Convert_to_pixel (x[j], x_scale, x_offset);
      thePoints[j].y = (short) Convert_to_pixel (u[j], u_scale, u_offset);
    }

  /* Set the color */
  Set_color (color);

  /* Draw to pixmap */
  XDrawLines (theDisplay, theUPixmap, theGC, thePoints, npoints,
	      CoordModeOrigin);

  free (thePoints);
}

/* ========================================================================= */

void
EZplot_points1 (double *x, double *u, int npoints, char color)
{
  /* Draw a set point points (x,u) to the pixmap */

  XPoint *thePoints;
  int j;

  if (npoints <= 1)
    return;

  /* Allocate thePoints and copy x and u to thePoints */
  thePoints = (XPoint *) malloc ((unsigned) (npoints) * sizeof (XPoint));
  for (j = 0; j < npoints; j++)
    {
      thePoints[j].x = (short) Convert_to_pixel (x[j], x_scale, x_offset);
      thePoints[j].y = (short) Convert_to_pixel (u[j], u_scale, u_offset);
    }

  /* Set the color */
  Set_color (color);

  /* Draw to pixmap */
  XDrawPoints (theDisplay, theUPixmap, theGC, thePoints, npoints,
	       CoordModeOrigin);

  free (thePoints);
}

/* ========================================================================= */

void
EZplot_clear (char color)
{
  /* Clear the plot area by drawing a (plot_size)x(plot_size) colored square
   * to the pixmap */

  Set_color (color);
  XFillRectangle (theDisplay, theUPixmap, theGC, 0, 0, plot_size, plot_size);
}

/* ========================================================================= */

void
EZplot_display (void)
{
  /* Copy the pixmap to the X-window, thereby making it visable. */

  XCopyArea (theDisplay, theUPixmap, theWindow, theGC, 0, 0, plot_size,
	     plot_size, plot_position[0], plot_position[1]);

  /* XSync necessary to keep display up to date. */
  XSync (theDisplay, (Bool) 0);
}

/* ========================================================================= */

static int
Convert_to_pixel (double x, double scale, double offset)
{
  /* Converts from user coordinate to pixel location within the plot region. */

  return ((int) (scale * (x - offset)));
}

/* ========================================================================= */
/* 		      Start Window Opening Code                              */
/* ========================================================================= */

void
EZplot_ini (double xmin, double xmax, double umin, double umax)
{
  /* Initialize the graphics. 

   * The plotting is done in a square of (plot_size)x(plot_size) pixels.
   * Convert_to_pixel() converts from (x,u) coordinates to pixel coordinates
   * within the square.  The square is displayed in a larger X-window to give
   * a nice bordering effect.  Hence the height and width of the X-window in
   * pixels are: height = width = plot_size + 2*BORDER.  The extra 2*BORDER
   * accounts for BORDER pixels around the plotted square.
   *
   * The array plot_position[2] controls where within the X-window the plot
   * appears.  The upper-left corner is copied to the X-window BORDER pixels
   * down and BORDER pixels to the right of the upper-left corner of the
   * X-window.  see EZplot_display() */

  int width, height;

  plot_size = PLOTSIZE;
  height = width = plot_size + 2 * BORDER;
  plot_position[0] = BORDER;
  plot_position[1] = BORDER;

  /* Used by Convert_to_pixel to convert from (x,u) to pixels */
  x_scale = (float) plot_size / (xmax - xmin);
  u_scale = (float) plot_size / (umin - umax);
  x_offset = xmin;
  u_offset = umax;

  /* Initialize X */
  InitX ();

  /* Open a window and call other necessary X routines */
  OpenWindow (WINX, WINY, width, height);

  /* Initialize colors */
  InitColors ();

  EZplot_clear ('w');
}

/* ========================================================================= */

static void
Set_color (char color)
{
  switch (color)
    {
    case 'r':
      XSetForeground (theDisplay, theGC, theRedPixel);
      break;
    case 'g':
      XSetForeground (theDisplay, theGC, theGreenPixel);
      break;
    case 'b':
      XSetForeground (theDisplay, theGC, theBluePixel);
      break;
    case 'w':
      XSetForeground (theDisplay, theGC, theWhitePixel);
      break;
    default:
      XSetForeground (theDisplay, theGC, theBlackPixel);
      break;
    }
}

/* ========================================================================= */

static void
InitX (void)
{
  /* Initializes X  */

  char *theDisplayName = NULL;

  /* Open the display */
  if ((theDisplay = XOpenDisplay (NULL)) == NULL)
    {
      fprintf (stderr,
	       "ERROR: Could not open a connection to X on display %s\n",
	       XDisplayName (theDisplayName));
      exit (1);
    }

  theScreen = DefaultScreen (theDisplay);
  theDepth = DefaultDepth (theDisplay, theScreen);
  thePWidth = DisplayWidth (theDisplay, theScreen);
  thePHeight = DisplayHeight (theDisplay, theScreen);
  theBlackPixel = BlackPixel (theDisplay, theScreen);
  theWhitePixel = WhitePixel (theDisplay, theScreen);
  theColormap = DefaultColormap (theDisplay, theScreen);

  if (theDepth == 1)
    {
      printf ("Warning: EZ_plot does not support monochrome displays \n");
      printf ("The library may or may not work for you \n");
    }

  /* Print useful information. I suggest printing this at this at least once */
  if (verbose > 1)
    {
      printf ("%s version %d of the X Window System, X%d R%d\n",
	      ServerVendor (theDisplay),
	      VendorRelease (theDisplay),
	      ProtocolVersion (theDisplay), ProtocolRevision (theDisplay));

      printf ("Color plane depth...........%d \n", theDepth);
      printf ("Display Width...............%d \n", thePWidth);
      printf ("Display Height..............%d \n", thePHeight);
      printf ("The display %s\n", XDisplayName (theDisplayName));
    }
}

/* ========================================================================= */

static void
OpenWindow (int winx, int winy, int width, int height)
{

  /* Creates and maps the X window.  Also creates the Graphics Context and
   * Pixmaps.  */

  XSizeHints theSizeHints;
  XTextProperty theWindowName, theIconName;
  char *Name = WINDOW_TITLE;
  unsigned long background;

  /* Set the background color (i.e. color of border around the square plots */
  background = BlackPixel (theDisplay, theScreen);
  if (!BACKGROUND)
    background = WhitePixel (theDisplay, theScreen);

  /* Create the window */
  theWindow = XCreateSimpleWindow (theDisplay,
				   RootWindow (theDisplay, theScreen),
				   winx, winy, width, height, 0,
				   background, background);

  /* Set standard window properties.  theWindowName and theIconName are set to
   * Name, which unless you change it, it will be EZ-Plot.  theSizeHints.flags
   * are set such that if winx and winy both positive, they try to locate the
   * window there, else let the window manager decide.  */

  XStringListToTextProperty (&Name, 1, &theWindowName);
  XStringListToTextProperty (&Name, 1, &theIconName);

  if (winx > 0 && winy > 0)
    {
      theSizeHints.flags = USPosition | PSize;
    }
  else
    {
      theSizeHints.flags = PPosition | PSize;
    }

  XSetWMProperties (theDisplay, theWindow, &theWindowName, &theIconName,
		    NULL, 0, &theSizeHints, NULL, NULL);

  /* This selects which events the program will see. */
  XSelectInput (theDisplay, theWindow, EVENT_MASK);

  /* Create the Graphics Context. */
  theGC = XCreateGC (theDisplay, theWindow, (unsigned long) 0, NULL);

  if (theGC == 0)
    {
      XDestroyWindow (theDisplay, theWindow);
      printf ("Could not create the GC \n");
      exit (1);
    }

  /* Now that the Graphics Context has been created, setting and changing most
   * graphics characteristics (such as line width or line type) can easily be
   * done with appropriate X routine calls.  I include an example of how to
   * set the line width to 2 pixels.  LineSolid has the obvious meaning.
   * CapRound and JoinRound are probably not important to you.
   *
   * XSetLineAttributes(theDisplay, theGC, (unsigned int) 2, LineSolid,
   *                    CapRound, JoinRound); */

  XSetLineAttributes (theDisplay, theGC, (unsigned int) LINE_WIDTH, LineSolid,
		      CapRound, JoinRound);

  /* Map the window, i.e. make it visible */
  XMapWindow (theDisplay, theWindow);
  XFlush (theDisplay);

  /* Create pixmap of size plot_size x plot_size */
  theUPixmap = XCreatePixmap (theDisplay, theWindow, plot_size, plot_size,
			      theDepth);
  if (theUPixmap == False)
    printf ("couldn't open UPixmap\n");
}

/* ========================================================================= */

static void
InitColors (void)
{
  /* Allocate Red, Green, Blue, pixels (black and white already allocated) */

  XColor theRGBColor, theHardwareColor;

  if (XAllocNamedColor (theDisplay, theColormap, "Red",
			&theRGBColor, &theHardwareColor))
    {
      theRedPixel = theHardwareColor.pixel;
    }
  else
    {
      printf ("Cannot allocate the Red Pixel \n");
    }

  if (XAllocNamedColor (theDisplay, theColormap, "Green",
			&theRGBColor, &theHardwareColor))
    {
      theGreenPixel = theHardwareColor.pixel;
    }
  else
    {
      printf ("Cannot allocate the Green Pixel \n");
    }

  if (XAllocNamedColor (theDisplay, theColormap, "Blue",
			&theRGBColor, &theHardwareColor))
    {
      theBluePixel = theHardwareColor.pixel;
    }
  else
    {
      printf ("Cannot allocate the Blue Pixel \n");
    }
}

/* ========================================================================= */

static void
QuitX (void)
{
  /* Free memory and close the theDisplay.  */

  XFreeGC (theDisplay, theGC);
  XFreePixmap (theDisplay, theUPixmap);
  XCloseDisplay (theDisplay);
}

/* ========================================================================= */
/*		      Start Event Handling Code                              */
/* ========================================================================= */

int
EZevent (void)
{
  static XEvent theEvent;
  static KeySym theKeySym;
  static int theKeyBufferMaxLen = 64;
  static char theKeyBuffer[65];
  static XComposeStatus theComposeStatus;
  int return_value = 0;

  /* X Event Loop

   * If there is something in the queue then each event is processed
   * in turn. When the queue is empty, and the mode (which may have
   * been changed by the events) is MODE_SIMULATING then control is
   * returned (presumably to the main loop in main()).  However, when
   * the queue is empty and the mode is not MODE_SIMULATING then
   * XNextEvent is called which blocks until the next X event is
   * received (eg. the space bar being pressed which sets the mode
   * back to MODE_SIMULATING and so control returns to main()). */

  while (XPending (theDisplay) || (ezmode != MODE_SIMULATING))
    {

      XNextEvent (theDisplay, &theEvent);

      switch (theEvent.type)
	{			/* switch according to X event */

	case KeyPress:		/* A KeyPress event has occurred. */

	  XLookupString ((XKeyEvent *) & theEvent, theKeyBuffer,
			 theKeyBufferMaxLen, &theKeySym, &theComposeStatus);

	  switch (theKeySym)
	    {			/* switch according to the pressed key */

	    case XK_P:
	    case XK_p:
	      EZ_Pause ();
	      break;

	    case XK_space:
	      EZ_Resume ();
	      break;

	    case XK_Q:
	    case XK_q:
	      return_value = EZ_Quit ();
	      break;

	    case XK_Escape:
	      EZ_Kill ();
	      break;

#ifdef MY_KEY_EVENT_HANDLE
	    default:
	      return_value =
		My_Key_Event_Handle (*XKeysymToString (theKeySym));
#endif

	    }			/* switch(theKeySym) */
	  break;

	case EnterNotify:
	  /* This case is necessary for window managers which do not
	   * set keyboard focus to theWindow when the pointer enters
	   * theWindow. */
	  XSetInputFocus (theDisplay, theWindow, RevertToPointerRoot,
			  CurrentTime);
	  break;

	case Expose:
	  /* Window mapped for the first time and if you uncover some
	   * part of the window. If you start paused and you see
	   * nothing in the window then its possible that the problem
	   * is that the first Expose event is not being caught for
	   * some reason. */
	  EZplot_display ();
	  break;

	}			/* switch (theEvent.type) */

    }				/* while (XPending(theDisplay) || 
				   (ezmode != MODE_SIMULATING)) */

  return (return_value);
}

/* ========================================================================= */

void
EZ_Pause (void)
{
  /* The simulation has paused -- go to MODE_VIEWING.  */

  if (ezmode == MODE_SIMULATING)
    {
      ezmode = MODE_VIEWING;
    }
}

/* ========================================================================= */

void
EZ_Resume (void)
{
  /* Return to MODE_SIMULATING. This has the effect of continuing the
     simulation via a return from Event_check(). */

  if (ezmode == MODE_VIEWING)
    {
      ezmode = MODE_SIMULATING;
    }
}

/* ========================================================================= */

int
EZ_Quit (void)
{
  /* Return to MODE_SIMULATING and return 1 */

  ezmode = MODE_SIMULATING;
  return (1);
}

/* ========================================================================= */

void
EZ_Kill (void)
{
  /* Hard Exit */

  exit (0);
}

/* ========================================================================= */
/*       		    Start 2D Plotting Code                           */
/* ========================================================================= */

void
EZplot2d_ini (double xmin, double xmax, double umin, double umax)
{
  EZplot_ini (xmin, xmax, xmin, xmax);
  c1 = umin / 2.;
  c2 = umax / 2.;
}

/* ========================================================================= */

void
EZplot_polar (double *u, double dr, double dtheta, int nr, int ntheta)
{
  int j1, j2;
  XPoint thePoints[4];
  int npoints = 4;
  int shape = Convex;
  int mode = CoordModeOrigin;
  double r, theta;

  for (j1 = 0; j1 < nr; j1++)
    {
      r = j1 * dr;
      for (j2 = 0; j2 < ntheta; j2++)
	{
	  theta = j2 * dtheta;
	  if (u[j1 * (ntheta + 1) + j2] > 0.1)
	    {
	      if (u[j1 * (ntheta + 1) + j2] > 0.9)
		{
		  Set_color ('r');
		}
	      else
		{
		  Set_color (' ');
		}
	    }
	  else
	    {
	      Set_color ('b');
	    }

	  thePoints[0].x = (short) Convert_to_pixel (r * cos (theta),
						     x_scale, x_offset);
	  thePoints[0].y = (short) Convert_to_pixel (r * sin (theta),
						     u_scale, u_offset);

	  thePoints[1].x = (short) Convert_to_pixel ((r + dr) * cos (theta),
						     x_scale, x_offset);
	  thePoints[1].y = (short) Convert_to_pixel ((r + dr) * sin (theta),
						     u_scale, u_offset);

	  thePoints[2].x =
	    (short) Convert_to_pixel ((r + dr) * cos (theta + dtheta),
				      x_scale, x_offset);
	  thePoints[2].y =
	    (short) Convert_to_pixel ((r + dr) * sin (theta + dtheta),
				      u_scale, u_offset);

	  thePoints[3].x = (short) Convert_to_pixel (r * cos (theta + dtheta),
						     x_scale, x_offset);
	  thePoints[3].y = (short) Convert_to_pixel (r * sin (theta + dtheta),
						     u_scale, u_offset);


	  XFillPolygon (theDisplay, theUPixmap, theGC, thePoints, npoints,
			shape, mode);
	}
    }
  EZplot_display ();
}

/* ========================================================================= */

void
EZplot_polar_line1 (double *r, double *theta, int nj, char color)
{
  /* Draw a colored curve (r,theta) to the pixmap */

  XPoint *thePoints;
  int j;

  if (nj <= 1)
    return;

  /* Allocate thePoints and copy x and u to thePoints */
  thePoints = (XPoint *) malloc ((unsigned) (nj) * sizeof (XPoint));
  for (j = 0; j < nj; j++)
    {
      thePoints[j].x = (short) Convert_to_pixel (r[j] * cos (theta[j]),
						 x_scale, x_offset);
      thePoints[j].y = (short) Convert_to_pixel (r[j] * sin (theta[j]),
						 u_scale, u_offset);
    }

  /* Set the color */
  Set_color (color);

  /* Draw to pixmap */
  XDrawLines (theDisplay, theUPixmap, theGC, thePoints, nj, CoordModeOrigin);
  EZplot_display ();

  free (thePoints);
}

/* ========================================================================= */

void
EZplot_simple2 (double *u, int nj)
{
  int j1, j2;
  short x, y;

  for (j1 = 0; j1 < nj; j1++)
    {
      for (j2 = 0; j2 < nj; j2++)
	{
	  if (u[j2 * nj + j1] > 0)
	    {
	      if (u[j2 * nj + j1] < c2)
		Set_color ('r');
	      else
		Set_color ('w');
	    }
	  else
	    {
	      if (u[j2 * nj + j1] > c1)
		Set_color ('b');
	      else
		Set_color ('g');
	    }

	  x = (short) Convert_to_pixel ((double) j1, x_scale, x_offset);
	  y = (short) Convert_to_pixel ((double) (j2 + 1), u_scale, u_offset);

	  XFillRectangle (theDisplay, theUPixmap, theGC, x, y,
			  x_scale + 1, -u_scale + 1);
	}
    }
  EZplot_display ();
}

/* ========================================================================= */

